import cache_functional_model
import cache_physical_model
import pathlib
from lxml import etree
import os
import time
import configparser


######  USER DEFINITIONS  ######

trace_files = [
    'all'
]
memory_hierarchy_configs = [
    'example.ini'
]

config_dir = './config_dir'
trace_dir =  './trace_dir'
result_dir = './result_dir'

do_functional_simulation = False
do_physical_simulation = False
do_combined_simulation = False
do_cacti_simulation = True
verbose = True

################################

if (do_functional_simulation):
    memory_hierarchies = []
    for config in memory_hierarchy_configs:
        memory_hierarchies.append(cache_functional_model.MemoryHierarchyModel(config_file="{}/{}".format(config_dir, config)))
    xml_roots = []
    for memory_hierarchy in memory_hierarchies:
        xml_roots.append(etree.Element(memory_hierarchy.model_name))

    if (verbose):
        print("\nRunning simulation with the following memory hierarchies:\n")
        for config in memory_hierarchy_configs:
            print(config)

    if (trace_files[0] == 'all'):
        trace_files = []
        for file in os.listdir(trace_dir):
            if file.endswith('.txt'):
                trace_files.append(file)

    if (verbose):
        print("\nRunning simulation with the following traces:\n")
        for trace in trace_files:
            print(trace)

    # RUN SIMULATION
    for trace_file in trace_files:
        if (verbose):
            print("\nStarting "+trace_file+"...")
            start_time = time.time()
        with open("{}/{}".format(trace_dir, trace_file), 'r') as fp:
            instruction = fp.readline()
            while instruction:
                address, write = instruction.strip().split(',')
                for memory_hierarchy in memory_hierarchies:
                    memory_hierarchy.memory_access(address=int(address), write=int(write))
                instruction = fp.readline()
        # Gather results in respective xml containers
        for i, memory_hierarchy in enumerate(memory_hierarchies):
            memory_hierarchy.write_results(xml_roots[i], trace_file)

        if (verbose):
            print("{} finished!\nExecution time: {} seconds\n\n".format(trace_file, (time.time() - start_time)))
        for memory_hierarchy in memory_hierarchies:
            memory_hierarchy.reset_scoreboards()

    # OUTPUT RESULTS TO FILES
    for i, memory_hierarchy in enumerate(memory_hierarchies):
        path_output = pathlib.Path('./{}/{}.xml'.format(result_dir, memory_hierarchy.model_name))
        with path_output.open('wb+') as fp:
            fp.write(etree.tostring(xml_roots[i], pretty_print=True))

if (do_physical_simulation):
    for config in memory_hierarchy_configs:
    config_parser = configparser.ConfigParser(inline_comment_prefixes=';') # Parser for cache_config.ini file
    config_file = "{}/{}".format(config_dir, config)
    config_parser.read(config_file)
    keys = config_parser.sections()
    cache_models = {}
    main_memory_models = {}
    for key in keys:
        if (key == 'general'):
            pass
        elif (config_parser[key]['memory_type'] == 'cache'):
            cache_models[key] = cache_physical_model.CachePhysicalModel(model_name=key, result_file="{}/python_results.csv".format(result_dir))
        else:
            main_memory_models[key] = cache_physical_model.MainMemoryPhysicalModel(model_name=key)

    for key in cache_models:
        cache_models[key].read_cache_properties(key, config_file)
        cache_models[key].calculate_and_output_results()
    for key in main_memory_models:
        main_memory_models[key].read_memory_properties(key, config_file)

if (do_combined_simulation):
    memory_hierarchies = []
    for config in memory_hierarchy_configs:
        memory_hierarchies.append(cache_functional_model.MemoryHierarchyModel(config_file="{}/{}".format(config_dir, config)))
    xml_roots = []
    for memory_hierarchy in memory_hierarchies:
        xml_roots.append(etree.Element(memory_hierarchy.model_name))

    if (verbose):
        print("\nRunning simulation with the following memory hierarchies:\n")
        for config in memory_hierarchy_configs:
            print(config)

    if (trace_files[0] == 'all'):
        trace_files = []
        for file in os.listdir(trace_dir):
            if file.endswith('.txt'):
                trace_files.append(file)

    if (verbose):
        print("\nRunning simulation with the following traces:\n")
        for trace in trace_files:
            print(trace)

    # RUN SIMULATION
    for trace_file in trace_files:
        if (verbose):
            print("\nStarting "+trace_file+"...")
            start_time = time.time()
        with open("{}/{}".format(trace_dir, trace_file), 'r') as fp:
            instruction = fp.readline()
            while instruction:
                address, write = instruction.strip().split(',')
                for memory_hierarchy in memory_hierarchies:
                    memory_hierarchy.memory_access(address=int(address), write=int(write))
                instruction = fp.readline()
        # Gather results in respective xml containers
        for i, memory_hierarchy in enumerate(memory_hierarchies):
            memory_hierarchy.write_results(xml_roots[i], trace_file)

        if (verbose):
            print("{} finished!\nExecution time: {} seconds\n\n".format(trace_file, (time.time() - start_time)))
        for memory_hierarchy in memory_hierarchies:
            memory_hierarchy.reset_scoreboards()

    # OUTPUT RESULTS TO FILES
    for i, memory_hierarchy in enumerate(memory_hierarchies):
        path_output = pathlib.Path('./{}/{}.xml'.format(result_dir, memory_hierarchy.model_name))
        with path_output.open('wb+') as fp:
            fp.write(etree.tostring(xml_roots[i], pretty_print=True))
    
    path_output = pathlib.Path('./{}/{combined_simulation_results.csv'.format(result_dir))
    fp = path_output.open('w+')
    fp.write('Memory Hierarchy,Benchmark,Execution Time (ns),Total Energy Usage (mJ), Total Leakage Power (uW)')

    for config in memory_hierarchy_configs:
        config_parser = configparser.ConfigParser(inline_comment_prefixes=';') # Parser for cache_config.ini file
        config_file = "{}/{}".format(config_dir, config)
        config_parser.read(config_file)
        keys = config_parser.sections()
        cache_models = {}
        main_memory_models = {}
        for key in keys:
            if (key == 'general'):
                pass
            elif (config_parser[key]['memory_type'] == 'cache'):
                cache_models[key] = cache_physical_model.CachePhysicalModel(model_name=key, result_file="{}/python_results.csv".format(result_dir))
            else:
                main_memory_models[key] = cache_physical_model.MainMemoryPhysicalModel(model_name=key)

        for key in cache_models:
            cache_models[key].read_cache_properties(key, config_file)
            cache_models[key].calculate_and_output_results()
        for key in main_memory_models:
            main_memory_models[key].read_memory_properties(key, config_file)

        tree = etree.parse('{}/{}.xml'.format(result_dir,config_parser['general']['hierarchy_name']))
        cache_config = tree.getroot()

        for trace in cache_config:
            execution_time = 0
            total_energy_usage = 0
            total_leakage = 0
            write_hits = 0
            write_misses = 0
            read_hits = 0
            read_misses = 0
            for memory_level in trace:
                for element in memory_level.iter():
                    if (element.tag == 'write_hits'):
                        write_hits = int(element.text)
                    elif (element.tag == 'write_misses'):
                        write_misses = int(element.text)
                    elif (element.tag == 'read_hits'):
                        read_hits = int(element.text)
                    elif (element.tag == 'read_misses'):
                        read_misses = int(element.text)
                if memory_level.tag in cache_models:
                    execution_time += (write_hits + write_misses + read_hits + read_misses) * cache_models[memory_level.tag].access_penalty
                elif memory_level.tag in main_memory_models:
                    execution_time += (write_hits + write_misses + read_hits + read_misses) * main_memory_models[memory_level.tag].access_penalty
            for memory_level in trace:
                for element in memory_level.iter():
                    if (element.tag == 'write_hits'):
                        write_hits = int(element.text)
                    elif (element.tag == 'write_misses'):
                        write_misses = int(element.text)
                    elif (element.tag == 'read_hits'):
                        read_hits = int(element.text)
                    elif (element.tag == 'read_misses'):
                        read_misses = int(element.text)
                if memory_level.tag in cache_models:
                    total_energy_usage += write_hits * (cache_models[memory_level.tag].write_hit_energy)
                    total_energy_usage += write_misses * (cache_models[memory_level.tag].write_miss_energy)
                    total_energy_usage += read_hits * (cache_models[memory_level.tag].read_hit_energy)
                    total_energy_usage += read_misses * (cache_models[memory_level.tag].read_miss_energy)
                    total_energy_usage += (cache_models[memory_level.tag].leakage_power * execution_time) / 1_000_000
                    total_leakage += cache_models[memory_level.tag].leakage_power

                elif memory_level.tag in main_memory_models:
                    total_energy_usage += write_hits * (main_memory_models[memory_level.tag].write_hit_energy)
                    total_energy_usage += write_misses * (main_memory_models[memory_level.tag].write_miss_energy)
                    total_energy_usage += read_hits * (main_memory_models[memory_level.tag].read_hit_energy)
                    total_energy_usage += read_misses * (main_memory_models[memory_level.tag].read_miss_energy)
                    total_energy_usage += (main_memory_models[memory_level.tag].leakage_power * execution_time) / 1_000_000
                    total_leakage += main_memory_models[memory_level.tag].leakage_power
            fp.write('{},{},{},{},{}'.format(cache_config.tag, trace.tag, execution_time, total_energy_usage, total_leakage))
        fp.close()
            

if (do_cacti):
    cacti_configs = {      # Read  Write  Leakage
        "cache_config"   : 	[0,      0,      0]
    }

    path_output = pathlib.Path('./{}/{cacti_simulation_results.csv'.format(result_dir))
    fp = path_output.open('w+')
    fp.write('Memory Hierarchy,Benchmark,Execution Time (ns),Total Energy Usage (mJ), Total Leakage Power (uW)')
    
    for config in memory_hierarchy_configs:
        config_parser = configparser.ConfigParser(inline_comment_prefixes=';') # Parser for cache_config.ini file
        config_file = "{}/{}".format(config_dir, config)
        config_parser.read(config_file)
        keys = config_parser.sections()
        cache_models = {}
        main_memory_models = {}
        for key in keys:
            if (key == 'general'):
                pass
            elif (config_parser[key]['memory_type'] == 'cache'):
                cache_models[key] = cache_physical_model.CachePhysicalModel(model_name=key, result_file="{}/python_results.csv".format(result_dir))
            else:
                main_memory_models[key] = cache_physical_model.MainMemoryPhysicalModel(model_name=key)

        for key in cache_models:
            cache_models[key].read_cache_properties(key, config_file)
            cache_models[key].calculate_and_output_results()
        for key in main_memory_models:
            main_memory_models[key].read_memory_properties(key, config_file)

        tree = etree.parse('{}/{}.xml'.format(result_dir,config_parser['general']['hierarchy_name']))
        cache_config = tree.getroot()

        cacti_configs[config_parser['general']['hierarchy_name']][0]

        for trace in cache_config:
            execution_time = 0
            total_energy_usage = 0
            write_hits = 0
            write_misses = 0
            read_hits = 0
            read_misses = 0
            for memory_level in trace:
                for element in memory_level.iter():
                    if (element.tag == 'write_hits'):
                        write_hits = int(element.text)
                    elif (element.tag == 'write_misses'):
                        write_misses = int(element.text)
                    elif (element.tag == 'read_hits'):
                        read_hits = int(element.text)
                    elif (element.tag == 'read_misses'):
                        read_misses = int(element.text)
                if memory_level.tag in cache_models:
                    execution_time += (write_hits + write_misses + read_hits + read_misses) * cache_models[memory_level.tag].access_penalty
                elif memory_level.tag in main_memory_models:
                    execution_time += (write_hits + write_misses + read_hits + read_misses) * main_memory_models[memory_level.tag].access_penalty

            cacti_configs[config_parser['general']['hierarchy_name']][0]

            for memory_level in trace:
                for element in memory_level.iter():
                    if (element.tag == 'write_hits'):
                        write_hits = int(element.text)
                    elif (element.tag == 'write_misses'):
                        write_misses = int(element.text)
                    elif (element.tag == 'read_hits'):
                        read_hits = int(element.text)
                    elif (element.tag == 'read_misses'):
                        read_misses = int(element.text)
                if memory_level.tag in cacti_configs:
                    total_energy_usage += write_hits * cacti_configs[config_parser['general']['hierarchy_name']][1]
                    total_energy_usage += write_misses * cacti_configs[config_parser['general']['hierarchy_name']][0]
                    total_energy_usage += read_hits * cacti_configs[config_parser['general']['hierarchy_name']][0]
                    total_energy_usage += read_misses * (cacti_configs[config_parser['general']['hierarchy_name']][0] + cacti_configs[config_parser['general']['hierarchy_name']][1] * 2)
                    total_energy_usage += (cacti_configs[config_parser['general']['hierarchy_name']][2] * execution_time) / 1_000_000

            fp.write('{},{},{},{},{}'.format(cache_config.tag, trace.tag, execution_time, total_energy_usage, total_leakage))
        fp.close()